#include "FreeRTOS.h"
#include "scheduler.h"
#include "scheduler_logic.h"
#include "semphr.h"
#include "wcrt.h"
#include "queue.h"

/* ========================================================================= */

//
// structures for tests
//

#if ( TIMING_TEST == 1 )
uint32_t cs_times[ TASK_COUNT ][ RELEASE_COUNT ];
int task_instance[ TASK_COUNT ] = { 0, 0, 0, 0 };
struct TaskInfo * task_running = NULL;
int task_instance_c = 0;
#endif

#if ( TIMING_TEST == 2 )
uint32_t ulNotifyCountValue = 0;
struct TaskInfo * task_running = NULL;
#endif

/* ========================================================================= */

#if ( configUSE_SCHEDULER_START_HOOK == 1 )
/* Application Hooks. */
extern void vSchedulerStartHook( void );
#endif

/* Callback function called from the FreeRTOS tick interrupt service. */
void vApplicationTickHook( void );

/* TPH function. */
static void prvSchedulerTaskScheduler( void * params );

/* Scheduler task handle. */
static TaskHandle_t xSchedulerTask = NULL;

/* Current periodic task. */
static struct TaskInfo * xCurrentTask = NULL;

#if ( USE_TASK_NOTIFY == 0 )
/* Binary semaphore for blocking/unblocking the scheduler task. */
static SemaphoreHandle_t xSchedSem;
#endif

/**
 * AppSched_Init()
 */
void vSchedulerInit( void )
{
	/* Create the scheduler task. */
	xTaskCreate( prvSchedulerTaskScheduler, NULL, 256, NULL, TASK_SCHEDULER_PRIORITY, &xSchedulerTask );

	/* Calculate the worst case response times for each task. */
	xWcrtCalculateTasksWcrt();

#if ( USE_TASK_NOTIFY == 0 )
	/* Creates the binary semaphore used to wake up and block the scheduler task. */
	xSchedSem = xSemaphoreCreateBinary();
#endif

	/* Call the application scheduler hook function. This function should be
	 * defined by the programmer of the application scheduler if needed.
	 */
	vSchedulerTaskSchedulerStartLogic();

#if ( configUSE_SCHEDULER_START_HOOK == 1 )
	/* Call the hook function. This function should be defined by the user. */
	vSchedulerStartHook();
#endif

	/* Start the FreeRTOS scheduler. */
	vTaskStartScheduler();

	for(;;);
}

void vSchedulerSetup( void )
{
	vSchedulerLogicSetup();
}

/**
 * AppSched_TaskCreate()
 */
BaseType_t xSchedulerTaskCreate( TaskFunction_t pxTaskCode, const char * const pcName, const uint16_t usStackDepth, void * const pvParameters, UBaseType_t uxPriority, struct TaskInfo ** pxCreatedTask, TickType_t xPeriod, TickType_t xDeadline, TickType_t xWcet )
{
	struct TaskInfo *pxTaskInfo = ( struct TaskInfo * ) pvPortMalloc( sizeof( struct TaskInfo ) );

	BaseType_t xRslt;

	if( pxTaskInfo != NULL )
	{
		/* Initialize the scheduler tasks TCBe members. */
		pxTaskInfo->xPriority = uxPriority;
		pxTaskInfo->xPeriod = xPeriod;
		pxTaskInfo->xDeadline = xDeadline;
		pxTaskInfo->xAbsolutDeadline = xDeadline;
		pxTaskInfo->xRelease = 0;
		pxTaskInfo->xWcet = xWcet;
		pxTaskInfo->xWcrt = 0;
		pxTaskInfo->uxReleaseCount = 0;
		pxTaskInfo->xCur = 0;
		pxTaskInfo->xFinished = 0;

		vSchedulerLogicAddTask( pxTaskInfo );

		/* Create the FreeRTOS task. */
		xRslt = xTaskCreate( pxTaskCode, pcName, usStackDepth, pxTaskInfo, TASK_PRIORITY, &( pxTaskInfo->xHandle ) );

		if( xRslt == pdPASS)
		{
			if( ( void * ) pxCreatedTask != NULL )
			{
				/* Pass the TCBe out. */
				*pxCreatedTask = pxTaskInfo;
			}

			/* The initial state of a app scheduled task is suspended. */
			vTaskSuspend( pxTaskInfo->xHandle );
		}
		else
		{
			vPortFree( pxTaskInfo );
		}
	}
	else
	{
		xRslt = pdFAIL;
	}

	return xRslt;
}

/**
 * AppSched_WaitForNextPeriod()
 */
void vSchedulerWaitForNextPeriod()
{
 	vTaskDelayUntil( &( xCurrentTask->xRelease ), xCurrentTask->xPeriod );
}

/**
 * Función invocada en cada interrupción del tick de reloj. Esta función se
 * ejecuta desde una ISR, por lo que debe ser breve, e invocar funciones de
 * FreeRTOS que terminen en "FromISR".
 *
 * AppSched_Tick()
 */
void vApplicationTickHook( void )
{
	/* Returns TRUE if the application scheduler task must be awakened. */
	BaseType_t result = vSchedulerTaskSchedulerTickLogic();

	if( result == pdTRUE )
	{
#if ( USE_TASK_NOTIFY == 0 )
		xSemaphoreGiveFromISR( xSchedSem, NULL );
		portYIELD_FROM_ISR( pdTRUE );
#else
		BaseType_t xHigherPriorityTaskWoken = pdFALSE;
		vTaskNotifyGiveFromISR( xSchedulerTask, &xHigherPriorityTaskWoken );
		portYIELD_FROM_ISR( xHigherPriorityTaskWoken );
#endif
	}
}

/**
 * Scheduler task body function.
 *
 * AppSched_TPH()
 */
static void prvSchedulerTaskScheduler( void* params )
{
    do
    {
#if ( TIMING_TEST == 1 )
		task_running = xCurrentTask;
#endif
#if ( TIMING_TEST == 2 )
		task_running = xCurrentTask;
#endif

        /* Scheduler logic */
		vSchedulerTaskSchedulerLogic( &xCurrentTask );

#if ( USE_TASK_NOTIFY == 0 )
        if( xSemaphoreTake( xSchedSem, portMAX_DELAY ) == pdFALSE )
        {
            break;
        }
#else
#if ( TIMING_TEST == 2 )
	    ulNotifyCountValue = 1;
#endif
        ulTaskNotifyTake( pdTRUE, portMAX_DELAY );
#endif
    }
    while( pdTRUE );

    /* If the tasks ever leaves the for cycle, kill it. */
    vTaskDelete( NULL );
}

/**
 * AppSched_ReleaseFromISR()
 */
void vSchedulerReleaseFromISR( struct TaskInfo * const pxTask )
{
	if( pxTask != NULL )
	{
		if( xSchedulerTask != xTaskGetCurrentTaskHandle() )
		{
			vSchedulerLogicAddTaskToReadyList( pxTask );
#if ( USE_TASK_NOTIFY == 0 )
			xSemaphoreGiveFromISR( xSchedSem, NULL );
			portYIELD_FROM_ISR( pdTRUE );
#else
			BaseType_t xHigherPriorityTaskWoken = pdFALSE;
			vTaskNotifyGiveFromISR( xSchedulerTask, &xHigherPriorityTaskWoken );
			portYIELD_FROM_ISR( xHigherPriorityTaskWoken );
#endif
		}
	}
}

/**
 * Funcion invocada por las macros traceTASK_DELAY y traceTASK_DELAY_UNTIL. Esto
 * indica que una instancia ha finalizado. Luego se realiza un UP del semaforo
 * xSchedSem, que despertara a la tarea planificadora.
 *
 * AppSched_Delay()
 */
extern void vSchedulerTaskDelay( void )
{
	if( xCurrentTask != NULL )
	{
		/* Wake up the scheduler task only if is an application scheduled task. */
		if( xCurrentTask->xHandle == xTaskGetCurrentTaskHandle() )
		{
			xCurrentTask->xFinished = 1;

			/* A vTaskDelayUntil() or vTaskDelay() invocation terminates the
			 * current release of the task. */
			vSchedulerLogicRemoveTaskFromReadyList( xCurrentTask );

			/* Wake up the scheduler task. */
#if ( USE_TASK_NOTIFY == 0 )
			xSemaphoreGive( xSchedSem );
#else
			xTaskNotifyGive( xSchedulerTask );
#endif
		}
	}
}

/**
 * Función invocada por las macros traceBLOCKING_ON_QUEUE_RECEIVE y
 * traceBLOCKING_ON_QUEUE_SEND.
 *
 * AppSched_Block()
 */
extern void vSchedulerTaskBlock( void* xResource )
{
#if ( USE_TASK_NOTIFY == 0 )
	/* First identify if the blocking task is not the scheduler task. */
	if( xSchedSem != ( QueueHandle_t ) xResource )
	{
		if( xCurrentTask != NULL )
		{
			/* Wake up the scheduler task only if is an application scheduled task. */
			if( xCurrentTask->xHandle == xTaskGetCurrentTaskHandle() )
			{
				vSchedulerLogicRemoveTaskFromReadyList( xCurrentTask );

				/* Wake up the scheduler task. */
				xSemaphoreGive( xSchedSem );
			}
		}
	}
#if ( TIMING_TEST == 1 )
	else
	{
		/* The TPH is blocked. */
		ulDelayTime1 = CPU_CYCLES - ulDelayTime;

		if( task_running != NULL )
		{
			if(task_instance[ task_running->xPriority ] < RELEASE_COUNT) {
				cs_times[ task_running->xPriority ][ task_instance[ task_running->xPriority ] ] = ulDelayTime1 - ulDelayTime;
				//cs_times[ task_running->xPriority ][ task_instance[ task_running->xPriority ] ] = xTaskGetTickCount();
				task_instance[ task_running->xPriority ] += 1; //= task_running->uxReleaseCount;
				task_instance_c += 1;
			}
		}
	}
#endif
#else
	if( xCurrentTask != NULL )
	{
		/* Wake up the scheduler task only if is an application scheduled task. */
		if( xCurrentTask->xHandle == xTaskGetCurrentTaskHandle() )
		{
			vSchedulerLogicRemoveTaskFromReadyList( xCurrentTask );

			/* Wake up the scheduler task. */
			xTaskNotifyGive( xSchedulerTask );
		}
	}
#endif
}

/**
 * We must identify if the invocation is called from the aperiodic
 * task, and not from the vSchedulerTaskSchedulerLogic().
 *
 * AppSched_Suspend()
 */
extern void vSchedulerTaskSuspend( void* xTask )
{
	if( xSchedulerTask != xTaskGetCurrentTaskHandle() )
	{
		if( xCurrentTask != NULL )
		{
			if( xCurrentTask->xHandle == ( TaskHandle_t ) xTask )
			{
				xCurrentTask->xFinished = 1;

				vSchedulerLogicRemoveTaskFromReadyList( xCurrentTask );

				/* Wake up the scheduler task. */
#if ( USE_TASK_NOTIFY == 0 )
				xSemaphoreGive( xSchedSem );
#else
				xTaskNotifyGive( xSchedulerTask );
#endif

			}
		}
	}
}

/**
 * Invoked by the traceMOVED_TASK_TO_READY_STATE macro. This macro is called by the
 * function prvAddTaskToReadyList() at:
 *
 * - xTaskGenericCreate()
 * - vTaskPrioritySet()
 * - vTaskResume()
 * - xTaskResumeFromISR() <-- Could be checked by traceTASK_RESUME_FROM_ISR(xTask).
 * - xTaskResumeAll()
 * - xTaskIncrementTick() <-- Could be checked by traceTASK_INCREMENT_TICK(xTickCount).
 * - xTaskRemoveFromEventList()
 * - xTaskRemoveFromUnorderedEventList()
 * - vTaskPriorityInherit()
 * - xTaskPriorityDisinherit()
 *
 * If the task is ready to execute, we must wake up the scheduler thread.
 *
 * If using timers, it requires that INCLUDE_xTimerGetTimerDaemonTaskHandle, so the
 * xTimerTaskHandle can be returned by the xTimerGetTimerDaemonTaskHandle() function.
 *
 * Maps to the AppSched_Ready() function in the HST model.
 */
extern void vSchedulerTaskReady( void* xTask )
{
    /*
	if( xSchedulerTask == ( TaskHandle_t ) xTask )
	{
		STOPWATCH_RESET();
		ulDelayTime = CPU_CYCLES;
	}
    */

	if( xSchedulerTask != xTaskGetCurrentTaskHandle() )
	{
		/* The scheduler task is not running. Then xTask had been moved to the
		 * ready task list by FreeRTOS, because it is a new release of a periodic
		 * task, or it has been unblocked.
		 */
		struct TaskInfo * pxTask = NULL;

#if( configUSE_TIMERS == 1 )
		/* Check if the unblocked task was the timer task. */
		if( xTimerGetTimerDaemonTaskHandle() == ( TaskHandle_t ) xTask )
		{
			pxTask = NULL;
		}
#else
		pxTask = xSchedulerGetTaskTCBe( xTask );
#endif

		if( pxTask != NULL )
		{
			if( pxTask->xFinished == 1 )
			{
				/* If pxTask is a new instance, update the absolute deadline of
				 * the release, reset the CPU counter and increment the release
				 * counter.
				 */
				pxTask->xAbsolutDeadline = pxTask->xRelease + pxTask->xDeadline;
				pxTask->uxReleaseCount = pxTask->uxReleaseCount + 1;
				pxTask->xFinished = 0;
				pxTask->xCur = 0;

				/* Add the task to the appropriate ready list. */
				vSchedulerLogicAddTaskToReadyList( pxTask );
                
                xFlag = 1;
			}

			/* Wake up the scheduler task. */
			if( xSchedulerTask != NULL )
			{
				/* If vSchedulerTaskReady is called from an ISR, we need to
				 * invoke the FromISR variant of xSemaphoreGive(). This can
				 * occur when the task is unblocked in xTaskIncrementTick().
				 */
#if ( USE_TASK_NOTIFY == 0 )
				xSemaphoreGiveFromISR( xSchedSem, NULL );
				portYIELD_FROM_ISR( pdTRUE );
#else
				BaseType_t xHigherPriorityTaskWoken = pdFALSE;
				vTaskNotifyGiveFromISR( xSchedulerTask, &xHigherPriorityTaskWoken );
				portYIELD_FROM_ISR( xHigherPriorityTaskWoken );
#endif
			}
		}
	}
}

#if ( TIMING_TEST == 2 )
void vMacroTaskSwitchedIn()
{
	if( ulNotifyCountValue == 1 )
	{
		ulDelayTime1 = CPU_CYCLES;
		if( ( task_running != NULL ) && ( task_running->xFinished == 1 ) )
		{
            /*
			if(task_instance[ task_running->xPriority ] < RELEASE_COUNT) {
				cs_costs[ task_running->xPriority ][ task_instance[ task_running->xPriority ] ] = ulDelayTime1 - ulDelayTime;
				//cs_times[ task_running->xPriority ][ task_instance[ task_running->xPriority ] ] = xTaskGetTickCount();
				task_instance[ task_running->xPriority ] += 1; //= task_running->uxReleaseCount;
				task_instance_c += 1;
			}
            */
            if( ( *cs_costs )[ task_running->xPriority ][ 0 ] < RELEASE_COUNT )
            {
                ( *cs_costs )[ task_running->xPriority ][ ( *cs_costs )[ task_running->xPriority][ 0 ] + 1 ] = ulDelayTime1 - ulDelayTime;
                ( *cs_costs )[ task_running->xPriority ][ 0 ] += 1;
            }
		}
		ulNotifyCountValue = 0;
	}
	else
	{
		STOPWATCH_RESET();
		ulDelayTime = CPU_CYCLES;
	}
}
#endif
