#include "FreeRTOS.h"
#include "scheduler.h"
#include "scheduler_logic.h"

#if ( USE_SLACK == 1 )
#include "slack.h"
#define MIN_SLACK 0
#endif

#define ONE_TICK ( ( TickType_t ) 1 )

int xFlag = 0;

extern void vSchedulerDeadlineMissHook( struct TaskInfo * xTask, const TickType_t xTickCount );

/* Periodic tasks list. */
List_t xBigTaskList;
List_t * pxBigTaskList = NULL;

/* Ready tasks list. */
static List_t xReadyTasksList;
static List_t * pxReadyTasksList = NULL;

#if ( USE_SLACK == 1 )
/* Flag that indicates if the system is using slack. */
static BaseType_t xUsingSlack = pdFALSE;

extern void vSchedulerNegativeSlackHook( TickType_t xTickCount, BaseType_t xSlack );

/* Aperiodic tasks list. */
static List_t xAllAperiodicTasksList;
static List_t * pxAllAperiodicTasksList = NULL;

/* Ready aperiodic / sporadic list. */
static List_t xAperiodicReadyTasksList;
static List_t * pxAperiodicReadyTasksList = NULL;
#endif

/**
 * AppSchedLogic_Init()
 */
void vSchedulerTaskSchedulerStartLogic( void )
{
	const ListItem_t * pxAppTasksListEndMarker = listGET_END_MARKER( pxBigTaskList );
	ListItem_t * pxAppTasksListItem = listGET_HEAD_ENTRY( pxBigTaskList );

	/* Initialize the ready task ready list. */
	vListInitialise( &( xReadyTasksList ) );
	pxReadyTasksList = &( xReadyTasksList );

#if ( USE_SLACK == 1 )
	/* Initialize the aperiodic task ready list. */
	vListInitialise( &( xAperiodicReadyTasksList ) );
	pxAperiodicReadyTasksList = &( xAperiodicReadyTasksList );
#endif

	/* Init periodic tasks. */
	while( pxAppTasksListEndMarker != pxAppTasksListItem )
	{
		/* Pointer to the application scheduled task. */
		struct TaskInfo * pxAppTask = ( struct TaskInfo * ) listGET_LIST_ITEM_OWNER( pxAppTasksListItem );

#if ( USE_SLACK == 1 )
		/* Init the slack structure. */
		struct TaskInfo_Slack * pxTaskInfoSlack = ( struct TaskInfo_Slack * ) pvPortMalloc( sizeof( struct TaskInfo_Slack ) );

		/* Initialize slack methods attributes. */
		pxTaskInfoSlack->xDi = 0;
		pxTaskInfoSlack->xSlack = 0;
		pxTaskInfoSlack->xTtma = 0;
		pxTaskInfoSlack->xK = 0;

		pxAppTask->vExt = pxTaskInfoSlack;

		vSlackCalculateSlack_fixed1( pxAppTask, 0 );

		pxTaskInfoSlack->xK = pxTaskInfoSlack->xSlack;
#endif
		/* Insert the task into the ready list. */
		vListInsert( pxReadyTasksList, &( pxAppTask->xReadyListItem ) );

		pxAppTasksListItem = listGET_NEXT( pxAppTasksListItem );
	}

#if ( USE_SLACK == 1 )
	/* Minimal slack at the critical instant. */
	vSlackUpdateAvailableSlack( &xAvailableSlack );
#endif
}

/**
 * AppSchedLogic_Tick()
 */
#if ( USE_SLACK == 1 )
BaseType_t vSchedulerTaskSchedulerTickLogic()
{
	BaseType_t xResult = pdFALSE;

	/* ==================   Update slack   ================== */

	if( xUsingSlack == pdTRUE )
	{
		/* Decrement all slack counters, and increment the current aperiodic
		 * task executed time. */
		vSlackDecrementAllTasksSlack( ONE_TICK );
		struct TaskInfo * pxAperiodicTask = ( struct TaskInfo * ) listGET_OWNER_OF_HEAD_ENTRY( pxAperiodicReadyTasksList );
		pxAperiodicTask->xCur = pxAperiodicTask->xCur + ONE_TICK;
	}
	else
	{
		if( listLIST_IS_EMPTY( pxReadyTasksList ) == pdFALSE )
		{
			/* Decrement higher priority tasks slack, and increment the current
			 * task executed time. */
			struct TaskInfo * pxTask = ( struct TaskInfo * ) listGET_OWNER_OF_HEAD_ENTRY( pxReadyTasksList );
			vSlackDecrementTasksSlack( pxTask , ONE_TICK );
		    pxTask->xCur = pxTask->xCur + ONE_TICK;
		}
		else
		{
			/* Idle task or system scheduled task. */
			vSlackDecrementAllTasksSlack( ONE_TICK );
		}
	}

	/* Update the xAvailableSlack global counter with the minimum slack. */
	vSlackUpdateAvailableSlack( &xAvailableSlack );

	if( xAvailableSlack < 0 )
	{
		vSchedulerNegativeSlackHook( xTaskGetTickCountFromISR(), xAvailableSlack );
	}

	/* ==================   Verify deadlines   ================== */

	const TickType_t xTickCount = xTaskGetTickCountFromISR();

	const ListItem_t * pxAppTasksListEndMarker = listGET_END_MARKER( pxReadyTasksList );
	ListItem_t * pxAppTasksListItem = listGET_HEAD_ENTRY( pxReadyTasksList );

	while( pxAppTasksListEndMarker != pxAppTasksListItem )
	{
		struct TaskInfo * xTask = ( struct TaskInfo * ) listGET_LIST_ITEM_OWNER( pxAppTasksListItem );

		if( xTask->xAbsolutDeadline < xTickCount )
		{
			vSchedulerDeadlineMissHook( ( struct TaskInfo * ) listGET_LIST_ITEM_OWNER( pxAppTasksListItem ), xTickCount );
		}

		pxAppTasksListItem = listGET_NEXT( pxAppTasksListItem );
	}

	/* ==================   Scheduler wake up   ================== */

	/* If the system slack is below the minimum, wake up the scheduler. */
	if( xUsingSlack == pdTRUE )
	{
		if( xAvailableSlack <= MIN_SLACK )
		{
			xResult = pdTRUE;
		}
	}

	return xResult;
}
#else
BaseType_t vSchedulerTaskSchedulerTickLogic()
{
	if( listLIST_IS_EMPTY( pxReadyTasksList ) == pdFALSE )
	{
		/* Increment the current task executed time. */
		struct TaskInfo * pxTask = ( struct TaskInfo * ) listGET_OWNER_OF_HEAD_ENTRY( pxReadyTasksList );
		pxTask->xCur = pxTask->xCur + ONE_TICK;
	}

	/* ==================   Verify deadlines   ================== */

	const TickType_t xTickCount = xTaskGetTickCountFromISR();

	const ListItem_t * pxAppTasksListEndMarker = listGET_END_MARKER( pxReadyTasksList );
	ListItem_t * pxAppTasksListItem = listGET_HEAD_ENTRY( pxReadyTasksList );

	while( pxAppTasksListEndMarker != pxAppTasksListItem )
	{
		struct TaskInfo * xTask = ( struct TaskInfo * ) listGET_LIST_ITEM_OWNER( pxAppTasksListItem );

		if( xTask->xAbsolutDeadline < xTickCount )
		{
			vSchedulerDeadlineMissHook( ( struct TaskInfo * ) listGET_LIST_ITEM_OWNER( pxAppTasksListItem ), xTickCount );
		}

		pxAppTasksListItem = listGET_NEXT( pxAppTasksListItem );
	}

	return pdFALSE;
}
#endif

/**
 * AppSchedLogic_Sched()
 */
#if ( USE_SLACK == 1 )
void vSchedulerTaskSchedulerLogic( struct TaskInfo **pxCurrentTask )
{
	/* Current RTOS tick value. */
	const TickType_t xTickCount = xTaskGetTickCount();

	const ListItem_t * pxAppTasksListEndMarker = listGET_END_MARKER( pxReadyTasksList );
    ListItem_t * pxAppTasksListItem = listGET_HEAD_ENTRY( pxReadyTasksList );

    /* Suspend all ready tasks. */
    if( xFlag == 1 )
    {
        while( pxAppTasksListEndMarker != pxAppTasksListItem )
        {
            struct TaskInfo * pxAppTask = ( struct TaskInfo * ) listGET_LIST_ITEM_OWNER( pxAppTasksListItem );

            if( eTaskGetState( pxAppTask->xHandle ) == eReady )
            {
                vTaskSuspend( pxAppTask->xHandle );
            }

            pxAppTasksListItem = listGET_NEXT( pxAppTasksListItem );
        }
        xFlag = 0;
    }

	/* Check if the current release of the periodic task has finished. */
	if( *pxCurrentTask != NULL )
	{
		if( ( *pxCurrentTask )->xPeriod > 0 )
		{
			if( listIS_CONTAINED_WITHIN( pxReadyTasksList, &( ( *pxCurrentTask )->xReadyListItem ) ) == pdFALSE )
			{
#if ( USE_SLACK_K == 0 )
				/* Recalculate slack. */
				vSlackCalculateSlack_fixed1( *pxCurrentTask, xTickCount );
#else
				struct TaskInfo_Slack * pxTaskSlack = ( struct TaskInfo_Slack * ) ( *pxCurrentTask )->vExt;
				pxTaskSlack->xSlack = pxTaskSlack->xK;
#endif

				if( ( *pxCurrentTask )->xWcet > ( *pxCurrentTask )->xCur )
				{
					/* The release executed for less than the task worst case
					 * execution time, so vSlackGainSlack() is called to assign
					 * this time to lower priority tasks.
					 */
					vSlackGainSlack( *pxCurrentTask, ( ( *pxCurrentTask )->xWcet - ( *pxCurrentTask )->xCur ) );
				}

				/* Update the available slack. */
				vSlackUpdateAvailableSlack( &xAvailableSlack );

				*pxCurrentTask = NULL;
			}
		}
	}

	if( xAvailableSlack < 0 )
	{
		vSchedulerNegativeSlackHook( xTickCount, xAvailableSlack );
	}

	if( xAvailableSlack > MIN_SLACK )
	{
		if( listLIST_IS_EMPTY( pxAperiodicReadyTasksList ) == pdFALSE )
		{
			if( xUsingSlack == pdFALSE )
			{
				/* No aperiodic tasks are running. */
				*pxCurrentTask = ( struct TaskInfo * ) listGET_OWNER_OF_HEAD_ENTRY( pxAperiodicReadyTasksList );
				vTaskResume( ( *pxCurrentTask )->xHandle );
				xUsingSlack = pdTRUE;
			}
		}
		else
		{
			xUsingSlack = pdFALSE;
		}
	}
	else
	{
		if( xUsingSlack == pdTRUE )
		{
			vTaskSuspend( ( *pxCurrentTask )->xHandle );
			xUsingSlack = pdFALSE;
		}
	}

	if( xUsingSlack == pdFALSE )
	{
	    /* Periodic task scheduling -- resume the execution of the first task
	     * in the ready list, if any. */
		if( listLIST_IS_EMPTY( pxReadyTasksList ) == pdFALSE )
		{
			*pxCurrentTask = ( struct TaskInfo * ) listGET_OWNER_OF_HEAD_ENTRY( pxReadyTasksList );
			vTaskResume( ( *pxCurrentTask )->xHandle );
		}
	}
}
#else
void vSchedulerTaskSchedulerLogic( struct TaskInfo **pxCurrentTask )
{
	const ListItem_t * pxAppTasksListEndMarker = listGET_END_MARKER( pxReadyTasksList );
    ListItem_t * pxAppTasksListItem = listGET_HEAD_ENTRY( pxReadyTasksList );

    /* Suspend all ready tasks. */
    if( xFlag == 1 )
    {
        while( pxAppTasksListEndMarker != pxAppTasksListItem )
        {
            struct TaskInfo * pxAppTask = ( struct TaskInfo * ) listGET_LIST_ITEM_OWNER( pxAppTasksListItem );

            if( eTaskGetState( pxAppTask->xHandle ) == eReady )
            {
                vTaskSuspend( pxAppTask->xHandle );
            }
            
            pxAppTasksListItem = listGET_NEXT( pxAppTasksListItem );
        }
        xFlag = 0;
    }

	/* Check if the current release of the periodic task has finished. */
	if( *pxCurrentTask != NULL )
	{
		if( ( *pxCurrentTask )->xPeriod > 0 )
		{
			if( listIS_CONTAINED_WITHIN( pxReadyTasksList, &( ( *pxCurrentTask )->xReadyListItem ) ) == pdFALSE )
			{
				*pxCurrentTask = NULL;
			}
		}
	}

	/* Periodic task scheduling -- resume the execution of the first task in the ready list, if any. */
	if( listLIST_IS_EMPTY( pxReadyTasksList ) == pdFALSE )
	{
		*pxCurrentTask = ( struct TaskInfo * ) listGET_OWNER_OF_HEAD_ENTRY( pxReadyTasksList );
		vTaskResume( ( *pxCurrentTask )->xHandle );
	}
}
#endif

/**
 *
 */
void vSchedulerLogicAddTaskToReadyList( struct TaskInfo *xTask )
{
	if( xTask->xPeriod > 0U )
	{
		vListInsert( pxReadyTasksList, &( xTask->xReadyListItem ) );
	}
#if ( USE_SLACK == 1 )
	else
	{
		vListInsert( pxAperiodicReadyTasksList, &( xTask->xReadyListItem ) );
	}
#endif
}

/**
 *
 */
void vSchedulerLogicRemoveTaskFromReadyList( struct TaskInfo *xTask )
{
	uxListRemove( &( xTask->xReadyListItem ) );
}

/**
 *
 */
struct TaskInfo* xSchedulerGetTaskTCBe( TaskHandle_t xTask )
{
	struct TaskInfo * pxTask = NULL;

	ListItem_t * pxAppTasksListItem = listGET_HEAD_ENTRY( pxBigTaskList );
	ListItem_t const * pxAppTasksListEndMarker = listGET_END_MARKER( pxBigTaskList );

	/* Check if xTask is a periodic task. */
	while( pxAppTasksListEndMarker != pxAppTasksListItem )
	{
		if( ( ( struct TaskInfo * ) listGET_LIST_ITEM_OWNER( pxAppTasksListItem ) )->xHandle == ( TaskHandle_t ) xTask )
		{
			pxTask = ( struct TaskInfo * ) listGET_LIST_ITEM_OWNER( pxAppTasksListItem );
			break;
		}
		pxAppTasksListItem = listGET_NEXT( pxAppTasksListItem );
	}

#if ( USE_SLACK == 1 )
	if( pxTask == NULL )
	{
		ListItem_t * pxAppTasksListItem = listGET_HEAD_ENTRY( pxAllAperiodicTasksList );
		ListItem_t const * pxAppTasksListEndMarker = listGET_END_MARKER( pxAllAperiodicTasksList );

		/* Check if xTask is an aperiodic task. */
		while( pxAppTasksListEndMarker != pxAppTasksListItem )
		{
			if( ( ( struct TaskInfo * ) listGET_LIST_ITEM_OWNER( pxAppTasksListItem ) )->xHandle == ( TaskHandle_t ) xTask )
			{
				pxTask = ( struct TaskInfo * ) listGET_LIST_ITEM_OWNER( pxAppTasksListItem );
				break;
			}
			pxAppTasksListItem = listGET_NEXT( pxAppTasksListItem );
		}
	}
#endif

	return pxTask;
}

/**
 *
 */
void vSchedulerLogicAddTask( struct TaskInfo * pxTask )
{
	/* Initialize the task's generic item list. */
	vListInitialiseItem( &( pxTask->xGenericListItem ) );
	listSET_LIST_ITEM_OWNER( &( pxTask->xGenericListItem ), pxTask );
	listSET_LIST_ITEM_VALUE( &( pxTask->xGenericListItem ), pxTask->xPriority );

	/* Initialize the task's ready item list. */
	vListInitialiseItem( &( pxTask->xReadyListItem ) );
	listSET_LIST_ITEM_OWNER( &( pxTask->xReadyListItem ), pxTask );
	listSET_LIST_ITEM_VALUE( &( pxTask->xReadyListItem ), pxTask->xPriority );

	/* Add the task to the correct appropiate task list. */
	if( pxTask->xPeriod > 0U )
	{
		/* Periodic task. */
		vListInsert( pxBigTaskList, &( pxTask->xGenericListItem ) );
	}
#if ( USE_SLACK == 1 )
	else
	{
		/* Aperiodic task. */
		vListInsert( pxAllAperiodicTasksList, &( pxTask->xGenericListItem ) );
	}
#endif
}

/**
 *
 */
void vSchedulerLogicSetup( void )
{
	/* Initialize the periodic task list. */
	vListInitialise( &( xBigTaskList ) );
	pxBigTaskList = &( xBigTaskList );

#if ( USE_SLACK == 1 )
	/* Initialize the aperiodic tasks list. */
	vListInitialise( &( xAllAperiodicTasksList ) );
	pxAllAperiodicTasksList = &( xAllAperiodicTasksList );
#endif
}
